#ifndef _DSTOUTE_CONFIG_H_
#define _DSTOUTE_CONFIG_H_
  #define UTE_LIBRARY_NAME "Ute"
  #define UTE_LIBRARY_DESCRIPTION "Ute - Utility Library"
  #define UTE_VERSION 0x010000
  #define UTE_VERSION_HEX "0x010000"
  #define UTE_VERSION_SHORT "1.0.0"
  #define UTE_VERSION_LONG "Ute-1.0.0"
  #define UTE_REVISION "1.0.0.911"
  #define UTE_REVISION_COUNT 911
  #define UTE_REVISION_INFO "d227b5f5d9ef2a4dc81e02c3e410756dcbf1b987"
#endif /* _DSTOUTE_CONFIG_H_ */
